"""Графический интерфейс для работы с сервером."""

import sys
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QApplication, QLabel, QTableView, QDialog, QPushButton, \
    QLineEdit, QFileDialog, QMessageBox, QHeaderView
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import Qt
import os


def create_users_table(database):
    list_users = [(user.id, user.login, user.info) for user in database.get_all_users()]
    users_table = QStandardItemModel()
    users_table.setVerticalHeaderLabels(['ID', 'Login', 'INFO'])
    for user in list_users:
        user_id, user_login, user_info = user
        user_id = QStandardItem(user_id)
        user_id.setEditable(False)
        user_login = QStandardItem(user_login)
        user_login.setEditable(False)
        user_info = QStandardItem(user_info)
        user_info.setEditable(False)
        users_table.appendRow([user_id, user_login, user_info])
    return users_table


def create_messages_table(database):
    msg_lst = database.get_all_messages()

    msg_table = QStandardItemModel()
    msg_table.setHorizontalHeaderLabels(['От кого', 'Кому', 'Сообщение', 'Время'])
    for msg in msg_lst:
        from_client, to_client, msg_text, time = msg
        from_client = QStandardItem(from_client)
        from_client.setEditable(False)
        to_client = QStandardItem(to_client)
        to_client.setEditable(False)
        msg_text = QStandardItem(msg_text)
        msg_text.setEditable(False)
        time = QStandardItem(str(time.replace(microsecond=0)))
        time.setEditable(False)
        msg_table.appendRow([from_client, to_client, msg_text, time])
    return msg_table


class MainWindow(QMainWindow):
    """Основное окно"""
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Кнопка выхода на тулбаре
        self.exit_action = QAction('Выход', self)
        self.exit_action.setShortcut('Ctrl+Q')
        self.exit_action.triggered.connect(qApp.quit)

        # Кнопка обновления списка клиентов на тулбаре
        self.refresh_action = QAction('Обновить список', self)

        # Кнопка для вывода истории сообщений на тулбаре
        self.msg_history_action = QAction('История сообщений', self)

        # Кнопка для настройки сервера на тулбаре
        self.config_action = QAction('Настройка сервера', self)

        # Статусбар
        self.statusBar()

        # Тулбар
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(self.exit_action)
        self.toolbar.addAction(self.refresh_action)
        self.toolbar.addAction(self.msg_history_action)
        self.toolbar.addAction(self.config_action)

        # Размеры основного окна
        self.setFixedSize(450, 400)
        self.setWindowTitle('Server Ivan Redkin')

        # Наименование таблицы
        self.table_title = QLabel('Списко всех клиентов', self)
        self.table_title.setFixedSize(400, 15)
        self.table_title.move(10, 35)

        # Таблица всех клиентов
        self.clients_table = QTableView(self)
        self.clients_table.setFixedSize(430, 300)
        self.clients_table.move(10, 55)

        # Отобразить окно
        self.show()



class HistoryMsgWindow(QDialog):
    """Окно с историей сообщений."""
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна
        self.setWindowTitle('История сообщений')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Кнопка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        # Таблица истории сообщений
        self.msg_history_table = QTableView(self)
        self.msg_history_table.move(10, 10)
        self.msg_history_table.setFixedSize(580, 620)

        # Отобразить окно
        self.show()



class ConfigWindow(QDialog):
    """Окно настроек."""

    def __init__(self):
        super(ConfigWindow, self).__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна
        self.setFixedSize(375, 240)
        self.setWindowTitle('Настройки')

        # Надпись 'Путь до файла БД'
        self.db_path_label = QLabel('Путь до файла БД: ', self)
        self.db_path_label.move(10, 10)
        self.db_path_label.setFixedSize(240, 15)

        # Отображение пути до файла БД
        self.db_path = QLineEdit(self)
        self.db_path.setFixedSize(250, 25)
        self.db_path.move(10, 30)
        self.db_path.setReadOnly(True)

        # Кнопка выбора пути
        self.db_path_button = QPushButton('Путь...', self)
        self.db_path_button.move(275, 28)

        # Функция обработчик кнопки выбора пути к файлу БД
        def open_file_bd():
            global dialog
            dialog = QFileDialog(self)
            path = dialog.getExistingDirectory()
            path = path.replace('/', '\\')
            self.db_path.insert(path)

        self.db_path_button.clicked.connect(open_file_bd)

        # Надпись 'Имя файла БД'
        self.db_name_label = QLabel('Имя файла БД: ', self)
        self.db_name_label.move(10, 68)
        self.db_name_label.setFixedSize(180, 27)

        # Отображение пути до файла БД
        self.db_name = QLineEdit(self)
        self.db_name.setFixedSize(257, 27)
        self.db_name.move(110, 68)

        # Надпись 'Порт'
        self.db_port_label = QLabel('Порт: ', self)
        self.db_port_label.move(10, 108)
        self.db_port_label.setFixedSize(180, 27)

        # Ввод номера порта
        self.db_name = QLineEdit(self)
        self.db_name.setFixedSize(257, 27)
        self.db_name.move(110, 108)

        # Надпись 'IP'
        self.db_ip_label = QLabel('IP: ', self)
        self.db_ip_label.move(10, 148)
        self.db_ip_label.setFixedSize(180, 27)

        # Ввод IP
        self.db_ip = QLineEdit(self)
        self.db_ip.setFixedSize(257, 27)
        self.db_ip.move(110, 148)

        # Кнопка сохранения настроек
        self.save_btn = QPushButton('Сохранить', self)
        self.save_btn.move(10, 200)

        # # Кнопка сохранения настроек
        self.close_btn = QPushButton('Закрыть', self)
        self.close_btn.move(275, 200)
        self.close_btn.clicked.connect(self.close)

        self.show()


# if __name__ == '__main__':
    # -------------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------------
    # 1. Проверка работы основново окна.
    # app = QApplication(sys.argv)
    # main_window = MainWindow()
    # main_window.statusBar().showMessage('Текст для статусбара')
    # test_list = QStandardItemModel(main_window)
    # test_list.setHorizontalHeaderLabels(['ID', 'Login', 'INFO'])
    # test_list.appendRow([QStandardItem('1'), QStandardItem('test1'), QStandardItem('test1')])
    # test_list.appendRow([QStandardItem('2'), QStandardItem('test2'), QStandardItem('test2')])
    # test_list.appendRow([QStandardItem('3'), QStandardItem('test3'), QStandardItem('test3')])
    # main_window.clients_table.setModel(test_list)
    # main_window.clients_table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
    # main_window.clients_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
    # app.exec_()
    #
    # -------------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------------
    #
    # 2. Проверка работы окна историй сообщений.
    # app = QApplication(sys.argv)
    # history_window = HistoryMsgWindow()
    # test_list = QStandardItemModel(history_window)
    # test_list.setHorizontalHeaderLabels(['От кого', 'Кому', 'Сообщение', 'Время'])
    # test_list.appendRow(
    #     [QStandardItem('test1'), QStandardItem('test2'), QStandardItem('Привет'), QStandardItem('01.01.01 00:00:00')])
    # test_list.appendRow(
    #     [QStandardItem('test2'), QStandardItem('test3'), QStandardItem('Привет тебе'), QStandardItem('02.02.02 01:01:01')])
    # test_list.appendRow(
    #     [QStandardItem('test3'), QStandardItem('test1'), QStandardItem('Пока'), QStandardItem('03.03.03 02:02:02')])
    # history_window.msg_history_table.setModel(test_list)
    # history_window.msg_history_table.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)
    # history_window.msg_history_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeToContents)
    # app.exec_()
    #
    # -------------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------------
    #
    # 3. Проверка работы окна настроек.
    # app = QApplication(sys.argv)
    # dial = ConfigWindow()
    # app.exec_()



"""Управление базой данных. Разработал Редкин И.А. декабрь 2022 г."""


from sqlalchemy import create_engine, Table, Column, Integer, String, MetaData, ForeignKey, DateTime
from sqlalchemy.orm import mapper, sessionmaker, relationship
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime


class ServerStorage:
    """Основной класс управления базой данных. БД - sqlite. Декларативный стиль."""
    Base = declarative_base()

    def __init__(self):
        self.engine = create_engine('sqlite:///server_data_base.db3',
                                    echo=False,
                                    pool_recycle=7200)

        self.Base.metadata.create_all(self.engine)
        Session = sessionmaker(bind=self.engine)

        self.sess = Session()

    class ClientDB(Base):
        """Таблица клиентов."""
        __tablename__ = 'clients'
        id = Column(Integer, primary_key=True)
        login = Column(String, unique=True)
        info = Column(String, nullable=True)
        passwd_hash = Column(String)
        pubkey = Column(String)

        def __init__(self, login, info, passwd_hash, pubkey):
            self.login = login
            self.info = info
            self.passwd_hash = passwd_hash
            self.pubkey = pubkey

    class ClientHistoryDB(Base):
        """Таблица историй посещения клиентов."""
        __tablename__ = 'client_history'
        id = Column(Integer, primary_key=True)
        client = Column(ForeignKey('clients.id'))
        entry_time = Column(DateTime)
        ip_address = Column(String)

        def __init__(self, client, entry_time, ip_address):
            self.client = client
            self.entry_time = entry_time
            self.ip_address = ip_address

    class ClientContactDB(Base):
        """Таблица контактов клиентов."""
        __tablename__ = 'client_contacts'
        id = Column(Integer, primary_key=True)
        from_client = Column(ForeignKey('clients.id'))
        to_client = Column(ForeignKey('clients.id'))
        msg = Column(String)
        time = Column(DateTime)

        def __init__(self, from_client, to_client, msg, time):
            self.from_client = from_client
            self.to_client = to_client
            self.msg = msg
            self.time = time

    def add_user(self, username, passwd_hash, pubkey):
        """Добавить пользователя в таблицу клиентов.
        :param username:
        :param passwd_hash:
        :param pubkey:
        :return:
        """
        user = self.ClientDB(username, '', passwd_hash, pubkey)
        self.sess.add(user)
        self.sess.commit()

    def insert_in_contactdb(self, from_client, to_client, msg, time):
        """Добавить информацию о сообщении в таблицу контактов.
        :param from_client:
        :param to_client:
        :param msg:
        :param time:
        :return:
        """
        dt = datetime.strptime(time, '%Y-%m-%d %H:%M:%S.%f')
        from_client_id = self.sess.query(self.ClientDB).filter(self.ClientDB.login == from_client).first()
        to_client_id = self.sess.query(self.ClientDB).filter(self.ClientDB.login == to_client).first()
        contact = self.ClientContactDB(from_client_id.id, to_client_id.id, msg, dt)
        self.sess.add(contact)
        self.sess.commit()

    def del_in_contactdb(self, from_client, to_client):
        """Удалить информацию о сообщении из таблицы контактов.
        :param from_client:
        :param to_client:
        :return:
        """
        from_client_id = self.sess.query(self.ClientDB).filter(self.ClientDB.login == from_client).first()
        to_client_id = self.sess.query(self.ClientDB).filter(self.ClientDB.login == to_client).first()
        r = self.sess.query(self.ClientContactDB).filter(self.ClientContactDB.from_client == from_client_id.id,
                                                         self.ClientContactDB.to_client == to_client_id.id
                                                         ).all()

        for el in r:
            self.sess.delete(el)
        self.sess.commit()

    def get_all_users_messages(self, username):
        """Получить все сообщения пользователя.
        :param username:
        :return:
        """
        user = self.sess.query(self.ClientDB).filter(self.ClientDB.login == username).first()
        r = self.sess.query(self.ClientContactDB).filter(
            (self.ClientContactDB.from_client == user.id) | (self.ClientContactDB.to_client == user.id)).all()
        result = []
        for el in r:
            from_client = self.sess.query(self.ClientDB).filter(self.ClientDB.id == el.from_client).first()
            to_client = self.sess.query(self.ClientDB).filter(self.ClientDB.id == el.to_client).first()

            result.append((from_client.login, to_client.login, el.msg, el.time))
        return result

    def get_all_users(self):
        """Получить всех пользователей сервера.
        :return:
        """
        return self.sess.query(self.ClientDB).all()

    def get_hash(self, username):
        """Получить хэш пароля в ascii.
        :param username:
        :return:
        """
        r = self.sess.query(self.ClientDB).filter(self.ClientDB.login == username).first()
        result = r.passwd_hash.encode('ascii')
        return result


    def get_all_contacts_login(self):
        """Получить все логины зарегистрированых клиентов.
        :return:
        """
        r = self.sess.query(self.ClientDB).all()
        all_contacts_login = [user.login for user in r]
        return set(all_contacts_login)

    def get_all_messages(self):
        """Получить все сообщения.
        :return:
        """
        all_messages_list = []
        all_messages = self.sess.query(self.ClientContactDB).all()
        for message in all_messages:
            from_client = self.sess.query(self.ClientDB).filter(self.ClientDB.id == message.from_client).first()
            to_client = self.sess.query(self.ClientDB).filter(self.ClientDB.id == message.to_client).first()
            all_messages_list.append((from_client.login, to_client.login, message.msg, message.time))
        return all_messages_list

    def get_my_contacts(self, username):
        """Получить все контакты пользователя.
        :param username:
        :return:
        """
        my_contacts_logins = []
        user = self.sess.query(self.ClientDB).filter(self.ClientDB.login == username).first()
        user_contacts_id = self.sess.query(self.ClientContactDB.to_client.distinct()).filter(
            self.ClientContactDB.from_client == user.id).all()
        for user_id in user_contacts_id:
            contact = self.sess.query(self.ClientDB).filter(self.ClientDB.id == user_id[0]).first()
            my_contacts_logins.append(contact.login)
        return my_contacts_logins

    def check_user(self, username):
        """Проверить наличе зарегистрированного пользователя.
        :param username:
        :return:
        """
        if self.sess.query(self.ClientDB).filter_by(login=username).count():
            return True
        return False


if __name__ == '__main__':
    s = ServerStorage()



"""Программа-сервер. Разработал Редкин И.А. декабрь 2022 г."""

import binascii
import dis
import hmac
import os
import sys
import threading
from socket import socket, AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR
import logging
import select
import logs.config_server_log
from common.variables import DEFAULT_PORT, MAX_CONNECTIONS, ACTION, TIME, \
    USER, ACCOUNT_NAME, FROM, PRESENCE, ERROR, MESSAGE, \
    MESSAGE_TEXT, RESPONSE_400, TO, RESPONSE_200, EXIT, GET_ALL_CLIENTS, RESPONSE_202, ALERT, GET_MY_CONTACTS, \
    ADD_CONTACT, CONTACT_NAME, AUTH, STATE, REG, PASSWORD_HASH, PUBLIC_KEY, RESPONSE_511, DATA, RESPONSE
from common.utils import get_msg, send_msg, log

from server_db import ServerStorage

SERVER_LOGGER = logging.getLogger('server')


class ServerVerifier(type):
    """Метакласс. Проверка на корректность класса сервера.
    Необходимо использовать сокеты для работы по TCP протоколу.
    """

    def __new__(cls, future_class_name, future_class_parents, future_class_attrs):

        lst_argval = []
        for func in future_class_attrs:
            try:
                ret = dis.get_instructions(future_class_attrs[func])
            except TypeError:
                pass
            else:
                for i in ret:
                    lst_argval.append(i.argval)
        if 'connect' in lst_argval:
            raise 'Вызов connect должен отсутствовать для сокетов сервера'
        if 'AF_INET' not in lst_argval or 'SOCK_STREAM' not in lst_argval:
            raise 'Необходимо использовать сокеты для работы по TCP протоколу'

        return type.__new__(cls, future_class_name, future_class_parents, future_class_attrs)


class PortDescriptor:
    """Класс-дескриптор. Корректность установки номера порта."""
    def __init__(self, default=DEFAULT_PORT):
        self.default = default

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__.get(self.my_attr, self.default)

    def __set__(self, instance, value):
        if value < 1024 or value > 65535:
            raise ValueError
        instance.__dict__[self.my_attr] = value

    def __delete__(self, instance):
        del instance.__dict__[self.my_attr]

    def __set_name__(self, owner, my_attr):
        self.my_attr = my_attr


@log
def arg_parser():
    """Разбор аргументов командной строки.
    -a - адрес;
    -p - порт;
    """
    listen_port = 7777
    listen_address = ''
    try:
        if '-p' in sys.argv:
            listen_port = int(sys.argv[sys.argv.index('-p') + 1])
    except IndexError:
        SERVER_LOGGER.error("При запуске клиента после параметра '-p' не указан номер порта.")
        exit()
    except ValueError:
        SERVER_LOGGER.error('При запуске клиента неверно указан номер порта.')
        sys.exit(1)

    try:
        if '-a' in sys.argv:
            listen_address = sys.argv[sys.argv.index('-a') + 1]
        else:
            listen_address = ''
    except IndexError:
        SERVER_LOGGER.error("При запуске клиента после параметра '-a' не указан адрес.")
        sys.exit(1)

    return listen_port, listen_address


class Server(metaclass=ServerVerifier):
    """Основной класс. Обменивается информацией с клиентами."""
    port = PortDescriptor()

    def __init__(self, listen_port, listen_address, database):
        super().__init__()
        self.db = database
        self.port = listen_port
        self.address = listen_address
        self.clients = []
        self.messages = []
        self.all_users = []
        self.names = dict()

    @log
    def process_client_message(self, message, messages_list, client, clients, names):
        """Разбирает словарь-сообщение от клиента."""
        SERVER_LOGGER.debug(f'Разбор сообщения от клиента : {message}')
        if ACTION in message and message[ACTION] == PRESENCE and \
                TIME in message and USER in message:
            # Если пользователь авторизуется
            if message[USER][STATE] == AUTH:
                if self.db.check_user(message[USER][ACCOUNT_NAME]):
                    # Процесс авторизации
                    SERVER_LOGGER.debug('Correct username, starting passwd check.')
                    message_auth = RESPONSE_511
                    random_str = binascii.hexlify(os.urandom(64))
                    message_auth[DATA] = random_str.decode('ascii')
                    hash = hmac.new(self.db.get_hash(message[USER][ACCOUNT_NAME]), random_str, 'MD5')
                    digest = hash.digest()
                    SERVER_LOGGER.debug(f'Auth message = {message_auth}')
                    try:
                        send_msg(client, message_auth)
                        ans = get_msg(client)
                    except OSError as err:
                        SERVER_LOGGER.debug('Error in auth, data:', exc_info=err)
                        client.close()
                    client_digest = binascii.a2b_base64(ans[DATA])
                    if RESPONSE in ans and ans[RESPONSE] == 511 and hmac.compare_digest(
                            digest, client_digest):
                        self.names[message[USER][ACCOUNT_NAME]] = client
                        return
                else:
                    self.response = RESPONSE_400
                    self.response[ERROR] = f'Пользователь {message[USER][ACCOUNT_NAME]} не зарегистрирован'
                    send_msg(client, self.response)
                    clients.remove(client)
                    client.close()
                return
            elif message[USER][STATE] == REG:
                if self.db.check_user(message[USER][ACCOUNT_NAME]):
                    self.response = RESPONSE_400
                    self.response[ERROR] = f'Имя {message[USER][ACCOUNT_NAME]} уже занято.'
                    send_msg(client, self.response)
                    clients.remove(client)
                    client.close()
                else:
                    self.db.add_user(message[USER][ACCOUNT_NAME],
                                     message[USER][PASSWORD_HASH],
                                     message[USER][PUBLIC_KEY])
                    clients.remove(client)
                    client.close()

        elif ACTION in message and message[ACTION] == MESSAGE and \
                TO in message and TIME in message \
                and FROM in message and MESSAGE_TEXT in message:
            self.db.insert_in_contactdb(message[FROM], message[TO], message[MESSAGE_TEXT], message[TIME])
            messages_list.append(message)
            return
        elif ACTION in message and message[ACTION] == GET_ALL_CLIENTS and ACCOUNT_NAME in message:
            self.response = RESPONSE_202
            for user in self.db.get_all_users():
                self.all_users.append(user.login)
            self.response[ALERT] = self.all_users
            send_msg(client, self.response)
            return
        elif ACTION in message and message[ACTION] == EXIT and ACCOUNT_NAME in message:
            clients.remove(names[message[ACCOUNT_NAME]])
            names[message[ACCOUNT_NAME]].close()
            del names[message[ACCOUNT_NAME]]
            return
        else:
            self.response = RESPONSE_400
            self.response[ERROR] = 'Запрос некорректен.'
            send_msg(client, self.response)
            return

    @log
    def process_message(self, message, names, listen_socks):
        """Отправляет сообщение от одного пользователя другому."""
        if message[TO] in names and names[message[TO]] in listen_socks:
            send_msg(names[message[TO]], message)
            SERVER_LOGGER.info(f'Отправлено сообщение пользователю {message[TO]} '
                               f'от пользователя {message[FROM]}.')
        elif message[TO] in names and names[message[TO]] not in listen_socks:
            raise ConnectionError
        else:
            SERVER_LOGGER.error(
                f'Пользователь {message[TO]} не зарегистрирован на сервере, '
                f'отправка сообщения невозможна.')

    def main(self):
        """Запуск сервера. Установка соединений с клиентами.
        Получение сообщений от клиентов. Отправка сообщения клиентам."""
        SERVER_LOGGER.info(
            f'Запущен сервер, порт для подключений: {self.port}, '
            f'адрес с которого принимаются подключения: {self.address}. '
            f'Если адрес не указан, принимаются соединения с любых адресов.')
        sock = socket(AF_INET, SOCK_STREAM)
        sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        sock.bind((self.address, self.port))
        sock.settimeout(0.5)

        self.sock = sock
        self.sock.listen()

        while True:
            try:
                client, addr = self.sock.accept()
            except OSError:
                pass
            else:
                SERVER_LOGGER.info(f'Установлено соедение с ПК {addr}')
                self.clients.append(client)

            self.r_clients = []
            self.w_clients = []

            try:
                if self.clients:
                    self.r_clients, self.w_clients, _ = select.select(self.clients, self.clients, [], 0)
            except OSError:
                pass

            if self.r_clients:
                for client_with_message in self.r_clients:
                    try:
                        self.process_client_message(get_msg(client_with_message),
                                                    self.messages, client_with_message, self.clients, self.names)
                    except Exception:
                        SERVER_LOGGER.info(f'Клиент {client_with_message.getpeername()} '
                                           f'отключился от сервера.')
                        self.clients.remove(client_with_message)

            # Если есть сообщения, обрабатываем каждое.
            for msg in self.messages:
                try:
                    self.process_message(msg, self.names, self.w_clients)
                    print('Сообщение отправлено')
                except Exception:
                    SERVER_LOGGER.info(f'Связь с клиентом с именем {msg[TO]} была потеряна')
                    self.clients.remove(self.names[msg[TO]])
                    del self.names[msg[FROM]]
            self.messages.clear()


if __name__ == '__main__':
    listen_port, listen_address = arg_parser()
    database = ServerStorage()

    s = Server(listen_port, listen_address, database)
    s.main()
